from ewah.hooks.base import EWAHBaseHook

from office365.runtime.auth.user_credential import UserCredential
from office365.sharepoint.client_context import ClientContext
from office365.sharepoint.files.file import File

from openpyxl import load_workbook

import io


class EWAHSharepointHook(EWAHBaseHook):
    _ATTR_RELABEL: dict = {
        "user": "login",
        "site_url": "schema",
    }

    conn_name_attr: str = "ewah_sharepoint_conn_id"
    default_conn_name: str = "ewah_sharepoint_default"
    conn_type: str = "ewah_sharepoint"
    hook_name: str = "EWAH Microsoft Sharepoint Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["port", "extra", "host"],
            "relabeling": {
                "password": "Password",
                "login": "User",
                "schema": "Sharepoint Site URL",
            },
        }

    def get_data_from_excel(
        self,
        relative_url: str,
        worksheet_name: str,
        header_row: int,
        start_row: int,
        batch_size: int,
    ):
        # adapted from: https://stackoverflow.com/a/69292234/14125255
        # (accessed 2021-10-15)

        assert (
            isinstance(batch_size, int) and batch_size > 0
        ), f"type: {type(batch_size)}, value: {str(batch_size)}"
        assert (
            isinstance(header_row, int) and header_row > 0
        ), f"type: {type(header_row)}, value: {str(header_row)}"
        assert (
            isinstance(start_row, int) and start_row > 0
        ), f"type: {type(start_row)}, value: {str(start_row)}"
        assert start_row > header_row

        ctx = ClientContext(self.conn.site_url).with_credentials(
            UserCredential(self.conn.user, self.conn.password)
        )
        response = File.open_binary(ctx, relative_url)
        bytes_file_obj = io.BytesIO()
        bytes_file_obj.write(response.content)
        bytes_file_obj.seek(0)
        ws = load_workbook(bytes_file_obj, data_only=True)[worksheet_name]
        headers = {
            ws.cell(row=header_row, column=col).value: col
            for col in range(1, ws.max_column + 1)
            if ws.cell(row=header_row, column=col)
        }
        while start_row < ws.max_row + 1:
            yield [
                {k: ws.cell(row=row, column=v).value for k, v in headers.items()}
                for row in range(start_row, min(start_row + batch_size, ws.max_row + 1))
            ]
            start_row += batch_size



from airflow.operators.python import PythonOperator as PO
from airflow.operators.dummy import DummyOperator as DO
from airflow.models import BaseOperator
from airflow.sensors.sql import SqlSensor

from ewah.hooks.base import EWAHBaseHook
from ewah.constants import EWAHConstants as EC

from datetime import datetime, timedelta, timezone
from copy import deepcopy
import pytz
import re


class EWAHSqlSensor(SqlSensor):
    """Overwrite native SQL sensor to allow usage of ewah custom connection types."""

    def _get_hook(self):
        conn = EWAHBaseHook.get_connection(conn_id=self.conn_id)
        if not conn.conn_type.startswith("ewah"):
            raise Exception("Must use an appropriate EWAH custom connection type!")
        return conn.get_hook()


class PGO(BaseOperator):
    """Airflow operator to execute PostgreSQL statements.

    Cannot use airflow.providers.postgres.operators.postgres.PostgresOperator
        due to the a requirement to have an SSH tunnel option.

    Needs to be otherwise interchangeable with Airflow's PostgresOperator.
    """

    template_fields = ("sql",)
    template_ext = (".sql",)

    def __init__(self, sql, postgres_conn_id, parameters=None, *args, **kwargs):
        self.sql = sql
        self.postgres_conn_id = postgres_conn_id
        self.parameters = parameters
        super().__init__(*args, **kwargs)

    def execute(self, context):
        hook = EWAHBaseHook.get_hook_from_conn_id(self.postgres_conn_id)
        hook.execute(sql=self.sql, params=self.parameters, commit=True)
        hook.close()  # SSH tunnel does not close if hook is not closed first


def datetime_utcnow_with_tz():
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def airflow_datetime_adjustments(datetime_raw):
    if type(datetime_raw) == str:
        datetime_string = datetime_raw
        if "-" in datetime_string[10:]:
            tz_sign = "-"
        else:
            tz_sign = "+"
        datetime_strings = datetime_string.split(tz_sign)

        if "T" in datetime_string:
            format_string = "%Y-%m-%dT%H:%M:%S"
        else:
            format_string = "%Y-%m-%d %H:%M:%S"

        if "." in datetime_string:
            format_string += ".%f"

        if len(datetime_strings) == 2:
            if ":" in datetime_strings[1]:
                datetime_strings[1] = datetime_strings[1].replace(":", "")
            datetime_string = tz_sign.join(datetime_strings)
            format_string += "%z"
        elif "Z" in datetime_string:
            format_string += "Z"
        datetime_raw = datetime.strptime(
            datetime_string,
            format_string,
        )
    elif not (datetime_raw is None or isinstance(datetime_raw, datetime)):
        raise Exception(
            "Invalid datetime type supplied! Supply either string"
            + " or datetime.datetime! supplied: {0}".format(str(type(datetime_raw)))
        )

    if datetime_raw and datetime_raw.tzinfo is None:
        datetime_raw = datetime_raw.replace(tzinfo=timezone.utc)

    return datetime_raw



from ewah.hooks import connection_types


def get_provider_info():
    return {
        "package-name": "ewah",
        "name": "EWAH",
        "description": "ELT With Airflow Helper - EWAH. Make EL easy for the world.",
        "connection-types": connection_types,
        "version": "1.0.0",
    }



from ewah.hooks.airtable import EWAHAirtableHook

from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC


class EWAHAirtableOperator(EWAHBaseOperator):
    _NAMES = ["airtable"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
    }

    _CONN_TYPE = EWAHAirtableHook.conn_type

    def __init__(self, base_id, table_id, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.table_id = table_id
        self.base_id = base_id

    def ewah_execute(self, context):
        self.upload_data(self.source_hook.get_table_data(self.base_id, self.table_id))



from ewah.hooks.base import EWAHBaseHook

import mailchimp3

from typing import List, Dict, Any


class EWAHMailchimpHook(EWAHBaseHook):
    # API docs: https://mailchimp.com/developer/api/marketing/

    _ATTR_RELABEL = {
        "api_key": "password",
        "user": "login",
    }

    conn_name_attr = "ewah_mailchimp_conn_id"
    default_conn_name = "ewah_mailchimp_default"
    conn_type = "ewah_mailchimp"
    hook_name = "EWAH Mailchimp Connection"

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        return {
            "hidden_fields": ["port", "schema", "extra", "host"],
            "relabeling": {
                "login": "User",
                "password": "API Key",
            },
        }

    def get_data_in_batches(
        self, resource: str, batch_size: int = 10000
    ) -> List[Dict[str, Any]]:
        client = mailchimp3.MailChimp(mc_user=self.conn.user, mc_api=self.conn.api_key)
        mc_resource = getattr(client, resource)
        keepgoing = True
        offset = 0
        while keepgoing:
            data = mc_resource.all(count=batch_size, offset=offset)[resource]
            if data:
                yield data
                offset += batch_size
            else:
                keepgoing = False



from ewah.hooks.infigo import EWAHInfigoHook

from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC


class EWAHInfigoOperator(EWAHBaseOperator):
    _NAMES = ["infigo"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _CONN_TYPE = EWAHInfigoHook.conn_type

    def __init__(self, resource=None, *args, **kwargs):
        resource = resource or kwargs.get("target_table_name")
        super().__init__(*args, **kwargs)
        self.resource = resource

    def ewah_execute(self, context):
        self.upload_data(
            self.source_hook.get_data(
                resource=self.resource,
            )
        )



from ewah.hooks.base import EWAHBaseHook
from ewah.constants import EWAHConstants as EC

# from datetime import datetime, date, timedelta
import pendulum
import requests
import gzip
import json
import time


class EWAHAmazonAdsHook(EWAHBaseHook):
    """
    Implements the Amazon Ads API.
    """

    _ENDPOINTS_ADS_API = {  # TODO: validate in operator
        "NA": "https://advertising-api.amazon.com",
        "EU": "https://advertising-api-eu.amazon.com",
        "FE": "https://advertising-api-fe.amazon.com",
    }

    _ENDPOINTS_TOKENS = {
        "NA": "https://api.amazon.com/auth/o2/token",
        "EU": "https://api.amazon.co.uk/auth/o2/token",
        "FE": "https://api.amazon.co.jp/auth/o2/token",
    }

    _ADS_TYPES = ["sp", "sb", "sd"]  # TODO: validate in operator

    _ATTR_RELABEL = {}

    conn_name_attr = "ewah_amazon_ads_conn_id"
    default_conn_name = "ewah_amazon_ads_default"
    conn_type = "ewah_amazon_ads"
    hook_name = "EWAH Amazon Ads API Connection"

    @staticmethod
    def get_ui_field_behaviour():
        # Hide all and use custom fields only
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login", "password"],
            "relabeling": {},
        }

    @staticmethod
    def get_connection_form_widgets():
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import (
            BS3TextFieldWidget,
            BS3PasswordFieldWidget,
        )
        from wtforms import StringField

        return {
            "extra__ewah_amazon_ads__lwa_client_id": StringField(
                "AWS LWA Client ID", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_ads__lwa_client_secret": StringField(
                "AWS LWA CLient Secret", widget=BS3PasswordFieldWidget()
            ),
            "extra__ewah_amazon_ads__region": StringField(
                "Amazon Ads Region (one of: NA, EU, FE)", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_ads__refresh_token": StringField(
                "Refresh Token", widget=BS3PasswordFieldWidget()
            ),
        }

    @property
    def access_token(self):
        if (
            not hasattr(self, "_access_token")
            or pendulum.now() > self._access_token_expires_at
        ):
            self.log.info("Requesting new access token...")
            requested_at = pendulum.now()
            url = self._ENDPOINTS_TOKENS[self.conn.region]
            payload = {
                "grant_type": "refresh_token",
                "client_id": self.conn.lwa_client_id,
                "client_secret": self.conn.lwa_client_secret,
                "refresh_token": self.conn.refresh_token,
            }
            response = requests.post(url, data=payload)
            assert response.status_code == 200, response.text
            self._access_token = response.json()["access_token"]
            self._access_token_expires_at = requested_at.add(
                seconds=response.json()["expires_in"] * 0.85
            )
            self.log.info(
                "New Amazon oauth access token is valid until {0}...".format(
                    self._access_token_expires_at.isoformat()
                )
            )
        return self._access_token

    def _make_request_with_backoff(self, request_func, max_retries=5):
        """
        Make a request with exponential backoff for rate-limited endpoints. Intially we take the retry-after value, but if we get multiple 429s in a row, we add an additional exponential backoff.
        """
        attempt = 0
        consecutive_429s = 0  # Track consecutive 429 responses
        base_wait = 1  # Base wait time for exponential backoff

        while attempt <= max_retries:
            response = request_func()
            
            # If successful response, return it
            if response.status_code == 200:
                consecutive_429s = 0
                return response
                
            # If rate limited (429), retry with backoff
            if response.status_code == 429:
                if attempt == max_retries:
                    break
                
                consecutive_429s += 1
                retry_after = int(response.headers.get('Retry-After', 1))
                
                if consecutive_429s > 2:
                    additional_backoff = base_wait * (2 ** (consecutive_429s - 2))
                    wait_time = retry_after + additional_backoff
                    self.log.info(
                        f"Multiple 429s detected. Adding backoff of {additional_backoff}s "
                        f"to Retry-After value of {retry_after}s"
                    )
                else:
                    wait_time = retry_after

                self.log.info(
                    f"Request throttled. Retrying in {wait_time} seconds "
                    f"(attempt {attempt + 1}/{max_retries})"
                )
                time.sleep(wait_time)
                attempt += 1
                continue
            
            # For any other error code, fail immediately
            assert False, (
                f"Request failed with status code {response.status_code}. "
                f"Response: {response.text}"
            )

        # If we've exhausted all retries on 429s, raise the last error
        assert False, (
            f"Request failed after {max_retries} retries due to rate limiting. "
            f"Status code: {response.status_code}, "
            f"Response: {response.text}"
        )

    def get_profile_ids(self):
        url = "/".join([self._ENDPOINTS_ADS_API[self.conn.region], "v2", "profiles"])
        headers = {
            "Amazon-Advertising-API-ClientId": self.conn.lwa_client_id,
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
        }
        response = self._make_request_with_backoff(
            lambda: requests.get(url, headers=headers)
        )
        return response.json()

    def get_report(
        self,
        date,
        ads_type,
        report_type,
        profile_id,
        additional_params=None,
    ):
        # profile_id - the account id to be used
        # ads_type - sp, sb, sd for sponsored products, brands, or display
        # report_type - e.g. keywords, targets

        self.log.info("Requesting report creation...")
        url = "/".join(
            [
                self._ENDPOINTS_ADS_API[self.conn.region],
                "v2",
                ads_type,
                report_type,
                "report",
            ]
        )
        headers = {
            "Amazon-Advertising-API-ClientId": self.conn.lwa_client_id,
            "Authorization": f"Bearer {self.access_token}",
            "Amazon-Advertising-API-Scope": str(profile_id),
            "Content-Type": "application/json",
        }
        params = {
            "reportDate": date.isoformat().replace("-", ""),
            "metrics": "cost,clicks,keywordId,keywordText",  # TODO: make kwarg
            **(additional_params or {})
            # "segment": "query",  # TODO: make kwarg
        }
        self.log.info(f"Creating report at {url}")
        response = requests.post(url, data=json.dumps(params), headers=headers)
        assert response.status_code == 202, response.text
        report_id = response.json()["reportId"]

        # Loop for report status
        wait_for = 1
        while True:
            url = "/".join(
                [self._ENDPOINTS_ADS_API[self.conn.region], "v2", "reports", report_id]
            )
            self.log.info(f"Pinging report status at {url}")
            response = requests.get(url, headers=headers)
            assert response.status_code == 200, response.text
            report_status = response.json()["status"]
            if report_status == "SUCCESS":
                break
            if report_status == "FAILURE":
                raise Exception(
                    f"Report failure!\n\nFull status request response:"
                    f"\n\n{response.text}\n\n"
                )
            if not report_status == "IN_PROGRESS":
                raise Exception(f"Invalid report status: {report_status}")
            self.log.info(f"In progress. Trying again in {wait_for}s...")
            time.sleep(wait_for)
            wait_for *= 2

        # Download data
        url = "/".join(
            [
                self._ENDPOINTS_ADS_API[self.conn.region],
                "v2",
                "reports",
                report_id,
                "download",
            ]
        )
        self.log.info(f"Downloading report from {url}")
        response = requests.get(url, headers=headers)
        assert response.status_code == 200, response.text
        report_data = json.loads(gzip.decompress(response.content).decode())

        return report_data

    def get_report_v3(
        self,
        start_date,
        end_date,
        ad_product,
        report_type,
        profile_id,
        additional_params=None,
    ):
        # profile_id - the account id to be used
        # ad_product - e.g. SPONSORED_PRODUCTS (sp), for now only sp
        # report_type - e.g. spCampaign, spAdvertisedProduct
        # additional_params - contains configuration params for API v3 like groupBy, columns, filters, etc.

        self.log.info("Requesting report creation...")
        url = "/".join(
            [
                self._ENDPOINTS_ADS_API[self.conn.region],
                "reporting",
                "reports",
            ]
        )
        headers = {
            "Amazon-Advertising-API-ClientId": self.conn.lwa_client_id,
            "Authorization": f"Bearer {self.access_token}",
            "Amazon-Advertising-API-Scope": str(profile_id),
            "Content-Type": "application/json",
        }
        params = {
            "startDate": start_date.isoformat(),
            "endDate": end_date.isoformat(),
            "configuration": {
                "adProduct": ad_product,
                "reportTypeId": report_type,
                "timeUnit": "DAILY",
                "format": "GZIP_JSON",
                **(additional_params),
            },
        }
        self.log.info(f"Creating report at {url}")
        response = self._make_request_with_backoff(
            lambda: requests.post(url, data=json.dumps(params), headers=headers)
        )
        report_id = response.json()["reportId"]

        # Loop for report status
        wait_for = 15 # min wait time from experience
        while True:
            url = "/".join(
                [
                    self._ENDPOINTS_ADS_API[self.conn.region],
                    "reporting",
                    "reports",
                    report_id,
                ]
            )
            self.log.info(f"Pinging report status at {url}")
            # Update headers with fresh token on each request
            headers["Authorization"]=f"Bearer {self.access_token}"
            response = self._make_request_with_backoff(
                lambda: requests.get(url, headers=headers)
            )
            report_status = response.json()["status"]
            if report_status == "COMPLETED":
                break
            if report_status == "FAILURE":
                raise Exception(
                    f"Report failure!\n\nFull status request response:"
                    f"\n\n{response.text}\n\n"
                )
            if not report_status in ("PROCESSING", "PENDING"):
                raise Exception(f"Invalid report status: {report_status}")
            self.log.info(f"In progress. Trying again in {wait_for}s...")
            time.sleep(wait_for)
            wait_for *= 2

        # Download data
        download_url = response.json()["url"]
        self.log.info(f"Downloading report from {download_url}")
        download = self._make_request_with_backoff(
            lambda: requests.get(download_url)
        )
        report_data = json.loads(gzip.decompress(download.content).decode())

        return report_data

    def get_report_dsp(
        self,
        account_id,
        report_date_start,
        report_date_end,
        report_type,
        metrics=None,
        dimensions=None,
    ):
        self.log.info("Requesting report creation...")
        url = "/".join(
            [
                self._ENDPOINTS_ADS_API[self.conn.region],
                "accounts",
                account_id,
                "dsp",
                "reports",
            ]
        )
        headers = {
            "User-Agent": "EWAH",
            "Accept": "application/vnd.dspcreatereports.v3+json",
            "Amazon-Advertising-API-ClientId": self.conn.lwa_client_id,
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
        }
        body = {
            "startDate": report_date_start,
            "endDate": report_date_end,
            "format": "JSON",
            "type": report_type,
            "timeUnit": "DAILY",
        }
        if metrics:
            body["metrics"] = metrics
        if dimensions:
            body["dimensions"] = dimensions

        response = requests.post(url, headers=headers, json=body)
        assert response.status_code == 202, response.text
        report_id = response.json()["reportId"]

        # Wait until report is ready
        url = "/".join([url, report_id])
        headers["Accept"] = "application/vnd.dspgetreports.v3+json"
        # Loop for report status
        wait_for = 1
        while True:
            self.log.info(f"Pinging report status at {url}")
            response = requests.get(url, headers=headers)
            assert response.status_code == 200, response.text
            report_status = response.json()["status"]
            if report_status == "SUCCESS":
                break
            if report_status == "FAILURE":
                raise Exception(
                    f"Report failure!\n\nFull status request response:"
                    f"\n\n{response.text}\n\n"
                )
            if not report_status == "IN_PROGRESS":
                raise Exception(f"Invalid report status: {report_status}")
            self.log.info(f"In progress. Trying again in {wait_for}s...")
            time.sleep(wait_for)
            wait_for *= 2

        url = response.json()["location"]
        self.log.info("Downloading report...")
        response = requests.get(url)
        assert response.status_code == 200, response.text
        return json.loads(response.content.decode())



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC
from ewah.hooks.sharepoint import EWAHSharepointHook


class EWAHSharepointOperator(EWAHBaseOperator):
    _NAMES = ["sharepoint"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _CONN_TYPE = EWAHSharepointHook.conn_type

    def __init__(
        self,
        file_relative_path: str,
        sheet_name: str,
        header_row: int = 1,
        start_row: int = 2,
        batch_size: int = 10000,
        *args,
        **kwargs,
    ) -> None:
        assert file_relative_path.endswith((".xlsx", ".xlsm")), "Only Excel files for now!"
        self.file_relative_path = file_relative_path
        self.sheet_name = sheet_name
        self.header_row = header_row
        self.start_row = start_row
        self.batch_size = batch_size
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context: dict) -> None:
        self.log.info("Loading data...")
        for batch in self.source_hook.get_data_from_excel(
            self.file_relative_path,
            self.sheet_name,
            self.header_row,
            self.start_row,
            self.batch_size,
        ):
            self.upload_data(batch)



from ewah.operators.base import EWAHBaseOperator
from ewah.utils.python_utils import is_iterable_not_string
from ewah.constants import EWAHConstants as EC

from ewah.hooks.google_ads import EWAHGoogleAdsHook

from datetime import datetime, timedelta


class EWAHGoogleAdsOperator(EWAHBaseOperator):
    _NAMES = ["gads", "google_ads"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHGoogleAdsHook.conn_type

    def __init__(
        self,
        client_id,
        resource,  # name of the resource, e.g. ad_group
        fields,  # dict of fields, e.g. {'campaign': ['id', 'name']}
        metrics=None,  # list of metrics to retrieve - deprecated!
        conditions=None,  # list of conditions, e.g. ["ad_group.status = 'ENABLED'"]
        *args,
        **kwargs
    ):
        if metrics:
            fields["metrics"] = metrics

        kwargs["primary_key"] = list(
            set(
                [
                    "segments__" + col.replace(".", "__")
                    for col in fields.get("segments", [])
                ]
                + ["{0}__resource_name".format(resource)]
            )
        )

        if conditions and not is_iterable_not_string(conditions):
            raise Exception('Argument "conditions" must be a list!')

        self.fields = fields
        self.resource = resource
        self.client_id = str(client_id)
        self.conditions = conditions

        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = "segments__date"

        super().__init__(*args, **kwargs)

        if not self.extract_strategy == EC.ES_FULL_REFRESH:
            assert "date" in fields.get(
                "segments", []
            ), "If using non-full-refresh loading, must include segments.date!"
            if self.extract_strategy == EC.ES_SUBSEQUENT:
                assert (
                    self.load_data_from or self.reload_data_from
                ), "If using subsequent loading, must supply either 'load_data_from' or 'reload_data_from'!"

    def get_select_statement(self, dict_format, prefix=None):
        # create the list of fields for the SELECT statement
        if prefix is None:
            prefix = ""
        elif not prefix[-1] == ".":
            prefix += "."
        fields = []
        for key, value in dict_format.items():
            for item in value:
                if type(item) == dict:
                    fields += self.get_select_statement(item, prefix + key)
                else:
                    fields += [prefix + key + "." + item]
        return fields

    def ewah_execute(self, context):
        # Make sure the correct date conditions are set
        conditions = self.conditions or []
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
            if isinstance(data_from, str):
                data_from = datetime.strptime(data_from, "%Y-%m-%d").date()
            data_from = data_from - (self.load_data_from_relative or timedelta(days=0))
            data_until = (self.data_until or datetime.now()).date()
        else:
            data_from = self.data_from.date()
            data_until = (self.data_until or datetime.now()).date()

        if self.load_data_chunking_timedelta:
            batch_from = data_from
            while batch_from <= data_until:
                batch_until = min(
                    data_until, batch_from + self.load_data_chunking_timedelta
                )
                conditions.append(
                    "segments.date BETWEEN '{0}' AND '{1}'".format(
                        batch_from.isoformat(),
                        batch_until.isoformat(),
                    )
                )
                self.upload_data(
                    self.source_hook.get_data(
                        client_id=self.client_id,
                        fields=self.fields,
                        resource=self.resource,
                        conditions=conditions,
                    )
                )
                del conditions[-1:]  # Re-added in next iteration
                batch_from = batch_until + timedelta(days=1)
        else:
            conditions.append(
                "segments.date BETWEEN '{0}' AND '{1}'".format(
                    data_from.isoformat(),
                    data_until.isoformat(),
                )
            )
            self.upload_data(
                self.source_hook.get_data(
                    client_id=self.client_id,
                    fields=self.fields,
                    resource=self.resource,
                    conditions=conditions,
                )
            )



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.hooks.oracle import EWAHOracleSQLOperator


class EWAHOracleSQLOperator(EWAHSQLBaseOperator):
    _NAMES = ["oracle"]

    _SQL_BASE = 'SELECT\n{columns}\nFROM "{schema}"."{table}"\n'
    _SQL_BASE_SELECT = "SELECT * FROM (\n\n{select_sql}\n\n) t\nWHERE {{0}}"
    _SQL_COLUMN_QUOTE = '"'
    _SQL_PARAMS = ":{0}"

    _CONN_TYPE = EWAHOracleSQLOperator.conn_type



from ewah.hooks.sql_base import EWAHSQLBaseHook

import pymysql

from typing import Optional, List, Union


class EWAHMySQLHook(EWAHSQLBaseHook):
    _DEFAULT_PORT = 3306

    _ATTR_RELABEL = {
        "user": "login",
        "database": "schema",
    }

    conn_name_attr = "ewah_mysql_conn_id"
    default_conn_name = "ewah_mysql_default"
    conn_type = "ewah_mysql"
    hook_name = "EWAH MySQL Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra"],
            "relabeling": {
                "password": "Password",
                "login": "User",
                "schema": "Database",
                "host": "Hostname / IP",
                "port": "Port (default: 3306)",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from wtforms import StringField

        return {
            f"extra__ewah_mysql__ssh_conn_id": StringField(
                "SSH Connection ID (optional)",
                widget=BS3TextFieldWidget(),
            ),
        }

    def _get_db_conn(self):
        conn = pymysql.connect(
            host=self.local_bind_address[0],
            port=self.local_bind_address[1],
            user=self.conn.user,
            passwd=self.conn.password,
            database=self.conn.database,
        )
        conn.begin()  # Begin transaction
        return conn

    def _get_cursor(self):
        return self.dbconn.cursor(cursor=pymysql.cursors.SSCursor)

    def _get_dictcursor(self):
        return self.dbconn.cursor(cursor=pymysql.cursors.SSDictCursor)

    def execute(
        self, sql: str, params: Optional[dict] = None, commit: bool = False, cursor=None
    ) -> None:
        self.log.info(
            "Executing SQL:\n\n{0}\n\nWith params:\n{1}".format(
                sql,
                "\n".join(
                    [
                        "{0}: {1}".format(key, str(value))
                        for (key, value) in params.items()
                    ]
                )
                if params
                else "No params!",
            )
        )
        (cursor or self.cursor).execute(sql.strip(), args=params)
        if commit:
            self.commit()



from ewah.uploaders.base import EWAHBaseUploader
from ewah.constants import EWAHConstants as EC

from airflow.operators.dummy import DummyOperator

from tempfile import NamedTemporaryFile
from datetime import datetime
from decimal import Decimal
import gspread
import json
import os


def monkeypatch_values_update(func_to_call):
    def monkeypatch_func(range_name, params, body):
        values = body.get("values")
        if values:
            values = [
                [
                    float(v)
                    if isinstance(v, Decimal)
                    else (
                        v.strftime("%Y-%m-%d %H:%M:%S%z")
                        if isinstance(v, datetime)
                        else v
                    )
                    for v in outer
                ]
                for outer in values
            ]
            body["values"] = values
        return func_to_call(range_name, params=params, body=body)

    return monkeypatch_func


class EWAHGSheetsUploader(EWAHBaseUploader):
    """
    The Google Sheets Hook works a little bit different to other hooks.
        - it can only have 1 upload call per task execution
        - it deletes all data from the sheet and re-uploads it to the sheet
        - the target schema name is the workbook key
        - the target table name is the worksheet name
    """

    def __init__(self, *args, **kwargs):
        self._upload_call_count = 0  # Make sure there is only a single upload
        super().__init__(EC.DWH_ENGINE_GS, *args, **kwargs)

    @classmethod
    def get_schema_tasks(cls, dag, *args, **kwargs):
        # Dummy tasks - there are no schemas in Google Sheets
        return (
            DummyOperator(
                task_id="kickoff",
                dag=dag,
            ),
            DummyOperator(
                task_id="final",
                dag=dag,
            ),
        )

    def detect_and_apply_schema_changes(self, *args, **kwargs):
        # need to overwrite parent function, does nothing
        raise Exception("This function is not implemented for this DWHook!")

    def commit(self):
        self.log.info("Everything is auto-committed")

    def rollback(self):
        raise Exception("Cannot rollback Google Sheets target!")

    def close(self):
        self.log.info("Nothing to close")

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,  # unused but always given
        columns_definition,
        load_strategy,  # Must be LS_INSERT_REPLACE
        upload_call_count,  # Must be 1
        primary_key=None,  # must accept arg, but it must also always be []
    ):
        # Google Sheets only works with drop & replace in one go
        assert (
            load_strategy == EC.LS_INSERT_REPLACE
        ), "Google Sheets DWHs can only be drop_and_replace!"
        assert upload_call_count == 1, "Chunking is not possible for Google Sheets DWH!"

        if primary_key:
            raise Exception("Arg primary_key invalidly supplied!")

        def colnum_string(n):
            # adapted from https://stackoverflow.com/questions/23861680/convert-spreadsheet-number-to-column-letter
            string = ""
            while n > 0:
                n, remainder = divmod(n - 1, 26)
                string = chr(65 + remainder) + string
            return string

        self.log.info("Replacing data in Google Sheets!")

        if not data:
            self.log.info("Nothing to upload!")
            return

        # authorize and get correct worksheet
        servie_acc_file = NamedTemporaryFile()
        self.log.info("Authenticating...")
        credentials = self.dwh_conn.extra_dejson
        credentials = credentials.get("client_secrets", credentials)
        service_acc = json.dumps(credentials)
        filename = os.path.abspath(servie_acc_file.name)
        with open(filename, "w") as f:
            f.write(service_acc)
        gclient = gspread.service_account(filename=filename)

        workbook = gclient.open_by_key(schema_name)
        worksheet = workbook.worksheet(table_name)
        values = worksheet.get_all_values()

        # Delete old data, if any existed
        if values:
            self.log.info("Deleting old data...")
            rows = len(values)
            columns = len(values[0])
            values = [columns * [""]] * rows
            range_notation = "A1:" + colnum_string(columns) + str(rows)
            worksheet.update(range_notation, values)  # delete!

        del values  # free up memory

        # insert new data - need to change the format! use columns definition
        self.log.info("Preparing data for Google Sheets upload...")
        column_header = list(columns_definition.keys())
        upload_data = [column_header]  # First row: headers
        while data:
            current_data = data.pop(0)
            upload_data.append([current_data.get(col, "") for col in column_header])
        range_notation = "A1:" + colnum_string(len(column_header))
        range_notation += str(len(upload_data))
        self.log.info("Uploading data now!")
        worksheet.spreadsheet.values_update = monkeypatch_values_update(
            worksheet.spreadsheet.values_update
        )
        worksheet.update(range_notation, upload_data)
        self.log.info("Upload done.")
